from logging import getLogger
from .data import Data, DataType
from io import StringIO
import json
from typing import TypeVar

logger = getLogger(__name__)
TRepository = TypeVar("Repository")


class JsonData(Data):
    def __init__(self, content: dict, repository: TRepository = None):
        super().__init__(repository, DataType.JSON)
        self.content_ = content

    @classmethod
    def load(cls, repository: TRepository) -> "JsonData":
        return repository.load(JsonData.create_from_json)

    @classmethod
    def create_from_json(cls, raw_data: bytes, repository: TRepository) -> "JsonData":
        sio = StringIO(raw_data.decode("utf-8"))
        content = json.load(sio)
        return JsonData(content, repository)

    @property
    def content(self) -> dict:
        return self.content_

    @property
    def __str__(self) -> str:
        return f"JsonData:{self.content}"



from logging import getLogger
from aksdp.task import Task
from aksdp.dataset import DataSet
from typing import List, Callable
from enum import Enum

logger = getLogger(__name__)


class TaskStatus(Enum):
    INIT = 0
    RUNNING = 1
    COMPLETED = 2
    ERROR = 3


class GraphTask(object):
    def __init__(self, task: Task, dependencies: List["GraphTask"]):
        """.ctor

        Args:
            task (Task): 実行タスク
            dependencies (List): 依存タスク
        """
        self.task = task
        self.status = TaskStatus.INIT
        self.input_ds = None
        self.output_ds = None
        self._pre_run_hook = self.empty_hook
        self._post_run_hook = self.empty_hook

        self._dependencies_static = dependencies if dependencies else []
        self._dependencies_dynamic = []

    @property
    def dependencies(self) -> List["GraphTask"]:
        return self._dependencies_static + self._dependencies_dynamic

    @property
    def dependencies_dynamic(self):
        return self._dependencies_dynamic

    @dependencies_dynamic.setter
    def dependencies_dynamic(self, tasks: List["GraphTask"]):
        self._dependencies_dynamic = tasks

    @property
    def pre_run_hook(self) -> Callable:
        """実行前フックの取得

        Returns:
            Callable: 実行前フック関数
        """
        return self._pre_run_hook

    @pre_run_hook.setter
    def pre_run_hook(self, v: Callable):
        """実行前フックの設定

        Args:
            v (Callable): 実行前フック関数
        """
        self._pre_run_hook = v

    @property
    def post_run_hook(self) -> Callable:
        """実行後フックの取得

        Returns:
            Callable: 実行後フック関数
        """
        return self._post_run_hook

    @post_run_hook.setter
    def post_run_hook(self, v: Callable):
        """実行後フックの設定

        Args:
            v (Callable): 実行後フック関数
        """
        self._post_run_hook = v

    def empty_hook(self, ds: DataSet):
        """デフォルトの空フック

        Args:
            ds (DataSet): データセット
        """
        pass

    def is_runnable(self) -> bool:
        """実行可能状態(未実行であり、かつ依存タスクが全て完了状態かどうか)

        Returns:
            bool: 実行可能状態
        """
        if self.status != TaskStatus.INIT:
            return False

        if self.dependencies is None:
            return True

        return all([gt.status == TaskStatus.COMPLETED for gt in self.dependencies])

    def run(self, ds: DataSet = None) -> "GraphTask":
        """タスク実行

        Args:
            ds (DataSet, optional): 入力DataSet. Defaults to None.

        Returns:
            GraphTask: 実行後のGraphTask
        """
        try:
            self.input_ds = ds
            self.status = TaskStatus.RUNNING
            logger.debug(f"task({self.task.__class__.__name__}) started.")
            logger.debug(f"  input_ds = {str(ds)}")
            self.pre_run_hook(ds)
            output_ds = self.task.gmain(ds)
            self.post_run_hook(output_ds)

            logger.debug(f"task({self.task.__class__.__name__}) completed. (elapse={self.task.elapsed_time:.3f}s)")
            logger.debug(f"  output_ds = {str(output_ds)}")
            self.output_ds = output_ds
            self.status = TaskStatus.COMPLETED
        except BaseException as e:
            logger.error(f"task({self.task.__class__.__name__}) failed. {str(e)}")
            self.status = TaskStatus.ERROR
            raise
        return self



from logging import getLogger
from aksdp.data import Data
from typing import List

logger = getLogger(__name__)


class DataSet:
    def __init__(self):
        self.data = {}

    def put(self, name: str, d: Data) -> "DataSet":
        self.data[name] = d
        return self

    def get(self, name: str) -> Data:
        return self.data[name]

    def keys(self) -> List[str]:
        return self.data.keys()

    def merge(self, ds: "DataSet") -> "DataSet":
        if ds:
            for k, v in ds.data.items():
                self.data[k] = v
        return self

    def save_all(self):
        for k, v in self.data.items():
            logger.debug(f"save dataset[{k}]...")
            v.save()

    def apply(self, task) -> "DataSet":
        return task.main(self)

    def __str__(self) -> str:
        keys = ",".join(self.keys())
        return f"DataSet({keys})"



import functools
from logging import getLogger
from pprint import pprint

from airflow import DAG
from airflow.operators.python_operator import PythonOperator
from aksdp.dataset import DataSet
from aksdp.graph import Graph

logger = getLogger(__name__)


def af_entrypoint(gt, ds, **kwargs):
    logger.info(f"ds={ds}")
    pprint(kwargs)

    # build Input DataSet from Xcoms
    ti = kwargs["ti"]
    ds = DataSet()
    for d in gt.dependencies:
        cn = d.task.__class__.__name__

        pull_ds = ti.xcom_pull(task_ids=cn)
        if pull_ds:
            logger.info(f"found Xcom from {cn}")
            ds.merge(pull_ds)

    # Run & retrun Xcoms
    return gt.task.main(ds)


class AirFlow(object):
    def __init__(self, g: Graph):
        self.graph = g
        self.dag_tasks = {}

    def to_dag(self, dag: DAG):
        logger.info("Task regist to dag...")
        # create operators
        for gt in self.graph.graph:
            cn = gt.task.__class__.__name__
            func = functools.partial(af_entrypoint, gt)
            op = PythonOperator(task_id=f"{cn}", provide_context=True, python_callable=func, dag=dag)
            logger.info(f"create airflow operator {cn}")
            self.dag_tasks[cn] = op

        # set dependencies
        for gt in self.graph.graph:
            cn = gt.task.__class__.__name__
            op = self.dag_tasks[cn]

            for dep in gt.dependencies:
                dep_cn = dep.task.__class__.__name__
                op_ups = self.dag_tasks[dep_cn]

                op.set_upstream(op_ups)
                logger.info(f"{cn} set_upstream to {dep_cn}")



from .repository import Repository as Repository
from .localfile_repository import LocalFileRepository as LocalFileRepository
from .pandas_db_repository import PandasDbRepository as PandasDbRepository
from .s3file_repository import S3FileRepository as S3FileRepository
from .sqlalchemy_repository import SqlAlchemyRepository as SqlAlchemyRepository



from aksdp.dataset import DataSet
from abc import ABCMeta, abstractmethod
from typing import List
import time


class Task(metaclass=ABCMeta):
    def __init__(self):
        """.ctor
        """
        self._in = {}
        self._elapsed_time = None

    @abstractmethod
    def main(self, d: DataSet) -> DataSet:
        """タスクの処理

        Args:
            d (DataSet): 入力DataSet

        Returns:
            DataSet: 出力DataSet
        """
        pass

    def gmain(self, d: DataSet) -> DataSet:
        """タスクの処理(Graphから呼び出す用)

        Args:
            d (DataSet): 入力DataSet

        Returns:
            DataSet: 出力DataSet
        """
        self._in = {}

        # DataSet 自動展開
        if d:
            for k in d.keys():
                self._in[k] = d.get(k).content

        # 実行
        start = time.time()
        r = self.main(d)
        self._elapsed_time = time.time() - start

        return r

    def input_datakeys(self) -> List[str]:
        """入力DataSetの中から実際に使用するデータの key を列挙する

        Returns:
            List[str]: 入力で使用するデータの key
        """
        return []

    def output_datakeys(self) -> List[str]:
        """出力DataSetに使用するデータのkeyを列挙する

        Returns:
            List[str]: 出力に使用するデータのkey
        """
        return []

    @property
    def elapsed_time(self):
        """実行時間の取得

        Returns:
            timedelta : 実行時間
        """
        return self._elapsed_time



from concurrent.futures import ThreadPoolExecutor
from logging import getLogger
from aksdp.graph import Graph, TaskStatus, GraphTask
from aksdp.dataset import DataSet
import time


logger = getLogger(__name__)


class ConcurrentGraph(Graph):
    def __init__(self, executor=None, catalog_ds: DataSet = DataSet(), disable_dynamic_dep: bool = False):
        super().__init__(catalog_ds, disable_dynamic_dep=disable_dynamic_dep)

        self.pool = executor
        if self.pool is None:
            self.pool = ThreadPoolExecutor()

    def _run(self, graph_task: GraphTask, input_ds: DataSet):
        # ProcessPoolを使用した場合、 GraphTask.run() 内でのアトリビュート更新が効かないので外から操作する
        graph_task.input_ds = input_ds
        graph_task.status = TaskStatus.RUNNING
        r = self.pool.submit(graph_task.run, input_ds)
        return r

    def run(self, ds: DataSet = None):
        last_ds = ds
        self.abort = False
        features = []

        while not self.abort:
            tasks = self.runnable_tasks()

            for t in tasks:
                input_ds = self._make_task_inputs(t, ds)
                features.append((t, self._run(t, input_ds)))

            # どれかの Feature が終わるまで待つ
            def all_feature_running(features):
                return all([f.running() for f in features])

            while all_feature_running([f[1] for f in features]):
                time.sleep(0.1)

            # 終了した features の result を吸い上げ、features を更新
            _next_featrues = []
            for f in features:
                if f[1].done():
                    try:
                        # 新しく返った GraphTask で差し替え
                        rgt = f[1].result()
                        self.graph[self.graph.index(f[0])] = rgt
                        last_ds = rgt.output_ds
                    except BaseException as e:
                        if not self._handle_error(f[0], f[0].input_ds, e):
                            raise
                else:
                    _next_featrues.append(f)
            features = _next_featrues

            if not features and not self.runnable_tasks():
                logger.debug("no runnables tasks, exit")
                break

        return last_ds



from logging import getLogger
from aksdp.data import Data, DataType
from io import StringIO
import pandas as pd
from .repository import Repository

logger = getLogger(__name__)


class PandasDbRepository(Repository):
    def __init__(self, engine, table_name):
        super().__init__()
        self.engine = engine
        self.table_name = table_name

    def save(self, data: Data):
        if data.data_type == DataType.DATAFRAME:
            data.content.to_sql(self.table_name, self.engine, if_exists="replace", index=False)
        else:
            raise ValueError("PandasDbRepository.save() not support DataType {data.data_type}")

    def load(self, ctor):
        sio = StringIO()
        df = pd.read_sql_table(self.table_name, self.engine)
        df.to_csv(sio, index=False)
        return ctor(sio.getvalue().encode("utf-8"), self)



from abc import ABCMeta, abstractmethod
from enum import Enum
from logging import getLogger
from typing import Generic, TypeVar, Any

logger = getLogger(__name__)
TRepository = TypeVar("Repository")


class DataType(Enum):
    """DataType定義
    """

    RAW = 0
    DATAFRAME = 1
    JSON = 2
    SQLALCHEMY_MODEL = 3


class Data(metaclass=ABCMeta):
    def __init__(self, repository: TRepository, data_type: DataType = DataType.RAW):
        """.ctor

        Args:
            repository (TRepository): リポジトリ
            data_type (DataType, optional): データの種類. Defaults to DataType.RAW.
        """
        self.repository = repository
        self.data_type_ = data_type

    def save(self):
        """保存
        """
        if self.repository:
            self.repository.save(self)

    @property
    def data_type(self) -> DataType:
        """DataType取得

        Returns:
            DataType: DataType
        """
        return self.data_type_

    @property
    def content(self) -> Any:
        """データの中身へのアクセス

        Returns:
            Any: データの中身
        """
        return None



from logging import getLogger
from aksdp.task import Task
from aksdp.dataset import DataSet
from typing import List, Callable
from .graph_task import GraphTask, TaskStatus

logger = getLogger(__name__)


class Graph:
    def __init__(self, catalog_ds: DataSet = None, disable_dynamic_dep: bool = False):
        """.ctor
        """
        self.graph = []
        self.error_handlers = []
        self.abort = False
        self.catalog_ds = catalog_ds if catalog_ds else DataSet()
        self.disable_dynamic_dep = disable_dynamic_dep

    def append(self, task: Task, dependencies: List[GraphTask] = []) -> GraphTask:
        """Taskの追加

        Args:
            task (Task): Taskインスタンス
            dependencies (List[GraphTask], optional): 依存タスク. Defaults to [].

        Returns:
            GraphTask: GraphTask
        """
        gt = GraphTask(task, dependencies)
        self.graph.append(gt)
        return gt

    def add_error_handler(self, cls, fn: Callable):
        """エラーハンドラの追加

        Args:
            cls (class): 例外の型
            fn (function): エラーハンドラのCallable
        """
        self.error_handlers.append((cls, fn))

    def _run(self, graph_task: GraphTask, input_ds: DataSet) -> DataSet:
        """タスクの起動

        Args:
            graph_task (GraphTask): 起動するタスク
            input_ds (DataSet): 入力DataSet

        Returns:
            [DataSet]: 出力DataSet
        """

        try:
            rt = graph_task.run(input_ds)
            return rt.output_ds
        except BaseException as e:
            if not self._handle_error(graph_task, input_ds, e):
                raise

    def _handle_error(self, graph_task: GraphTask, input_ds: DataSet, e: BaseException) -> bool:
        """エラー時の処理。
        対応するエラーハンドラが存在すればエラーハンドラ呼び出し

        Args:
            graph_task (GraphTask): エラーが発生したタスク
            input_ds (DataSet): 入力データセット
            e (BaseException): 例外インスタンス

        Returns:
            bool: エラーハンドリングしたかどうか
        """
        graph_task.status = TaskStatus.ERROR
        for eh in self.error_handlers:
            if isinstance(e, eh[0]):
                self.abort = True
                eh[1](e, input_ds)
                return True

        return False

    def _make_task_inputs(self, graph_task: GraphTask, default_ds: DataSet) -> DataSet:
        """タスクの入力DataSet作成

        Args:
            graph_task (Task): GraphTask
            default_ds (DataSet): デフォルトDataSet

        Returns:
            DataSet: 入力DataSet
        """
        ds = DataSet()
        ds.merge(self.catalog_ds)

        if not graph_task.dependencies:
            ds.merge(default_ds)
        else:
            for d in graph_task.dependencies:
                ds.merge(d.output_ds)
        return ds

    def run(self, ds: DataSet = None) -> DataSet:
        last_ds = ds
        self.abort = False

        while not self.abort:
            if not self.runnable_tasks():
                break

            t = self.runnable_tasks()[0]

            input_ds = self._make_task_inputs(t, ds)
            last_ds = self._run(t, input_ds)

        return last_ds

    def runnable_tasks(self) -> List[GraphTask]:
        """実行可能タスクの取得

        Returns:
            List[GraphTask]: 実行可能タスク
        """
        if self.disable_dynamic_dep:
            # 動的依存解決無効
            return [g for g in self.graph if g.is_runnable()]

        # 動的依存解決
        runnable_tasks = []
        init_tasks = [g for g in self.graph if g.status == TaskStatus.INIT]
        completed_tasks = [g for g in self.graph if g.status == TaskStatus.COMPLETED]

        def _find_datakey_provider(key: str) -> List[GraphTask]:
            return [gt for gt in completed_tasks if key in gt.task.output_datakeys()]

        for gt in init_tasks:
            deps = set()
            for ik in gt.task.input_datakeys():
                if ik in self.catalog_ds.keys():
                    continue

                prv = _find_datakey_provider(ik)

                if len(prv) != 1:
                    # logger.debug(f"data({ik}) provide from {len(prv)} tasks.")
                    break

                deps.add(prv[0])
            else:
                gt.dependencies_dynamic = list(deps)

                if gt.is_runnable():
                    runnable_tasks.append(gt)

        return runnable_tasks

    def autoresolve_dependencies(self):
        """グラフ依存関係の自動解決
        """
        logger.info("run task dependencies auto resolver...")

        def _find_datakey_provider(key: str) -> List[GraphTask]:
            return [gt for gt in self.graph if key in gt.task.output_datakeys()]

        for gt in self.graph:
            deps = set()
            for ik in gt.task.input_datakeys():
                prv = _find_datakey_provider(ik)
                prv_str = ",".join([gt.task.__class__.__name__ for gt in prv])

                if len(prv) > 1:
                    msg = f"data({ik}) provide from multiple task({prv_str})"
                    logger.error(msg)
                    raise ValueError(msg)

                if len(prv) == 0 and ik not in self.catalog_ds.keys():
                    # input を提供するタスクが無く、カタログにも同キーが無い場合はエラー
                    msg = f"task provide data({ik}) not found."
                    logger.info(msg)

                logger.debug(f"{gt.task.__class__.__name__} : depend on {prv[0].task.__class__.__name__} by data({ik})")
                deps.add(prv[0])

            gt.dependencies_dynamic = list(deps)

        logger.info("task autoresolve completed.")



from .graph import Graph as Graph
from .graph_task import TaskStatus as TaskStatus
from .graph_task import GraphTask as GraphTask
from .concurrent_graph import ConcurrentGraph as ConcurrentGraph
from .debug_graph import DebugGraph as DebugGraph

__all__ = [Graph, TaskStatus, GraphTask, ConcurrentGraph, DebugGraph]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aksdp',
 'aksdp.data',
 'aksdp.dataset',
 'aksdp.graph',
 'aksdp.repository',
 'aksdp.task',
 'aksdp.util']

package_data = \
{'': ['*']}

install_requires = \
['boto3>=1.13.25,<2.0.0', 'pandas>=1.0.4,<2.0.0', 'sqlalchemy>=1.3.17,<2.0.0']

setup_kwargs = {
    'name': 'aksdp',
    'version': '0.0.1.post3',
    'description': 'Simple DataPipeline Library',
    'long_description': '# aksdp\n\n## Overview\n\nA simple framework for writing data pipelines in Python\n\n## INSTALL\n\n```bash\n$ pip install aksdp\n```\n\n## QuickStart\n\n```python\n\nclass TaskA(Task):\n   def main(self, ds):\n      return ds\n\nclass TaskB(Task):\n   ...\n\nclass TaskC(Task):\n   ...\n\nclass TaskD(Task):\n   ...\n\ngraph = Graph()\ntask_a = graph.append(TaskA())\ntask_b = graph.append(TaskB(), [task_a])\ntask_c = graph.append(TaskC(), [task_b])\ntask_d = graph.append(TaskD(), [task_b, task_c])\ngraph.run()\n\n```\n\nEach task runs after each dependent task completes.  \nAlso, the data output upstream can be received as input data and processed.',
    'author': 'Y.Arimitsu',
    'author_email': 'yoshikazu_arimitsu@albert2005.co.jp',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/YoshikazuArimitsu/aksdp',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



from aksdp.graph import Graph, GraphTask
from aksdp.data import DataType
from aksdp.dataset import DataSet
from aksdp.repository import LocalFileRepository
from pathlib import Path
from logging import getLogger
import shutil
import pickle
import os

logger = getLogger(__name__)


class DebugGraph(Graph):
    """全タスクに実行前/後フックを差し込んで入出力を保存するGraph
    """

    def __init__(self, base_dir: Path, catalog_ds: DataSet = DataSet()):
        super().__init__(catalog_ds)
        logger.info(f"set up DebugGraph, base_dir={base_dir.name}")
        self.base_dir = base_dir

    @classmethod
    def save_ds_hook(cls, base_dir: Path, task_name: str):
        def dump_dataset(_base_dir, _task_name, ds):
            shutil.rmtree(_base_dir, ignore_errors=True)
            os.makedirs(_base_dir, exist_ok=True)

            # pickle DataSet
            with open(_base_dir / f"{_task_name}.pkl", "wb") as f:
                pickle.dump(ds, f)

            # dump DataSet data
            if ds:
                for k in ds.keys():
                    # 一瞬 Repository を差し替えてダンプ先に保存する
                    d = ds.get(k)
                    original_repo = d.repository

                    ext = ""
                    if d.data_type == DataType.DATAFRAME:
                        ext = ".csv"
                    if d.data_type == DataType.JSON:
                        ext = ".json"

                    dump_repo = LocalFileRepository(_base_dir / (k + ext))
                    d.repository = dump_repo

                    try:
                        d.save()
                    except BaseException as e:
                        logger.warning(f"dump failed, {str(e)}")

                    d.repository = original_repo

        def _save_ds(ds):
            dump_dataset(base_dir, task_name, ds)

        return _save_ds

    def append(self, task, dependencies=None):
        gt = super().append(task, dependencies)
        cn = task.__class__.__name__
        gt.pre_run_hook = DebugGraph.save_ds_hook(self.base_dir / Path(cn) / "in", cn)
        gt.post_run_hook = DebugGraph.save_ds_hook(self.base_dir / Path(cn) / "out", cn)
        return gt

    def run_task(self, gt: GraphTask):
        cn = gt.task.__class__.__name__
        path = self.base_dir / Path(cn) / "in" / f"{cn}.pkl"

        try:
            with open(path, "rb") as f:
                pkl = pickle.load(f)
        except BaseException as e:
            logger.error(f"can't load pickle, {str(e)}")
            return

        return gt.task.gmain(pkl)
